#Author(s): Bertram Mueller-Myhsok (Professor) and Tony Kam-Thong (PhD Candidate)
#Research Institute: Max Planck of Psychiatry
#Date: October 2009

#The purpose of this R source code is to compute the differences of correlation coefficients between Controls and Cases
#as a mean to isolate significant SNPs Interactions using gpuCor function of the gputools package on a CUDA enabled #graphic card. The evaluations are done in partitions in respect to the memory size of the graphic card.


#Input(s):
#	Inputmatrix	: First n offset column vectors consist of Subject Information.  Last column consists of the 
#			  phenotype and the remaining contains the SNPs in column vectors   
#	chunk		: The SNPs size of each partition
#	zscorethres	: Based on the desired retention rate
#	filename	: Location of where the results are progressively stored  
#	noff		: First n columns of subject information. Last column contains the phenotype (Cases=2, Controls=1) 
#Output(s):
#	Results are stored in a file specified under 'filename' where the partition IDs, diff_mpierence, Zscore and number #	of observations are stored.  
#	The first ten results along with the number of results for each partition tested will be written out on R console #	screen. 
library(gputools);  

#Inputs
total_index<-Inputmatrix@gtdata@nids;
chunk <- as.numeric(chunksize);
zthres<-4;
pthres<-as.numeric(pval_thres);
noff<-0;

#Stores cases and controls indices
cases_mpi <- which(Inputmatrix@phdata$pheno==2);
controls_mpi <- which(Inputmatrix@phdata$pheno==1);


# Determine the number of SNPs, should be the same
bigind <- Inputmatrix@gtdata@nsnps;

# determine number of splits (nsplits) by chunk -> this includes the incomplete one;
ceiling((bigind-noff)/chunk) -> nsplits;


cat('namesSet1',"\t",Inputmatrix@gtdata@snpnames[1],"\n",file=filename,append=T);

# Go into calculations;
# first do the complete ones
for ( i in 1:(nsplits-1)) {
        Inputmatrix_r<-as.double.snp.data(Inputmatrix@gtdata[,(((i-1)*chunk+1)+noff):((i*chunk)+noff)])
	for ( j in i:(nsplits-1)) {
		Inputmatrix_l<-as.double.snp.data(Inputmatrix@gtdata[,(((j-1)*chunk+1)+noff):((j*chunk)+noff)])
		casres_mpi <-gpuCor(Inputmatrix_r[cases_mpi,],Inputmatrix_l[cases_mpi,],use="pairwise.complete.obs",method="pearson");
		conres_mpi <- gpuCor(Inputmatrix_r[controls_mpi,],Inputmatrix_l[controls_mpi,],use="pairwise.complete.obs",method="pearson");
		diff_mpi <- casres_mpi$coeff-conres_mpi$coeff;
		zscore_mpi<-(diff_mpi-mean(as.vector(diff_mpi)))/sd(as.vector(diff_mpi));
		n_obs_mpi<-casres_mpi$pairs+conres_mpi$pairs;

		poverall<-2*(1-pnorm(abs(zscore_mpi)))
		results<-which(poverall<=pthres)
		
		cat(i,"\t",j,"\t",results,"\n",file=filename,append=T);
		cat(i,"\t",j,"\t",diff_mpi[results],"\n",file=filename,append=T);
		cat(i,"\t",j,"\t",zscore_mpi[results],"\n",file=filename,append=T);
		cat(i,"\t",j,"\t",poverall[results],"\n",file=filename,append=T);
		cat(i,"\t",j,"\t",n_obs_mpi[results],"\n",file=filename,append=T);
		cat(i,"\t",j,"\t",zscore_mpi[results[1:10]],'(',length(results),")\n",sep='');
      }
}

### still missing those last incomplete ones - eg i = 1 and j complete
j<-(nsplits-1)
Inputmatrix_l<-as.double.snp.data(Inputmatrix@gtdata[,((j*chunk+1)+noff):bigind]);
for ( i in 1:(nsplits-1)) {
		Inputmatrix_r<-as.double.snp.data(Inputmatrix@gtdata[,(((i-1)*chunk+1)+noff):((i*chunk)+noff)]);
		casres_mpi <- gpuCor(Inputmatrix_r[cases_mpi,],Inputmatrix_l[cases_mpi,],use="pairwise.complete.obs",method="pearson");
		conres_mpi <- gpuCor(Inputmatrix_r[controls_mpi,],Inputmatrix_l[controls_mpi,],use="pairwise.complete.obs",method="pearson");
		diff_mpi <- casres_mpi$coeff-conres_mpi$coeff;
		zscore_mpi<-(diff_mpi-mean(as.vector(diff_mpi)))/sd(as.vector(diff_mpi));
		n_obs_mpi<-casres_mpi$pairs+conres_mpi$pairs;

		poverall<-2*(1-pnorm(abs(zscore_mpi)))
		results<-which(poverall<=pthres)

		cat(i,"\t",j+1,"\t",results,"\n",file=filename,append=T);
		cat(i,"\t",j+1,"\t",diff_mpi[results],"\n",file=filename,append=T);
		cat(i,"\t",j+1,"\t",zscore_mpi[results],"\n",file=filename,append=T);
		cat(i,"\t",j+1,"\t",poverall[results],"\n",file=filename,append=T);
		cat(i,"\t",j+1,"\t",n_obs_mpi[results],"\n",file=filename,append=T);
		cat(i,"\t",j+1,"\t",zscore_mpi[results[1:10]],'(',length(results),")\n",sep='');
		    
}		
# now do the very last incomplete one

i<-(nsplits-1)
	j<-(nsplits-1)
		Inputmatrix_r<-as.double.snp.data(Inputmatrix@gtdata[,((i*chunk+1)+noff):bigind])
		Inputmatrix_l<-as.double.snp.data(Inputmatrix@gtdata[,((j*chunk+1)+noff):bigind])
		casres_mpi <- gpuCor(Inputmatrix_r[cases_mpi,],Inputmatrix_l[cases_mpi,],use="pairwise.complete.obs",method="pearson");
		conres_mpi <-gpuCor(Inputmatrix_r[controls_mpi,],Inputmatrix_l[controls_mpi,],use="pairwise.complete.obs",method="pearson");
		diff_mpi <- casres_mpi$coeff-conres_mpi$coeff;
		zscore_mpi<-(diff_mpi-mean(as.vector(diff_mpi)))/sd(as.vector(diff_mpi));
		n_obs_mpi<-casres_mpi$pairs+conres_mpi$pairs;
		
		poverall<-2*(1-pnorm(abs(zscore_mpi)))
		results<-which(poverall<=pthres)

		cat(i+1,"\t",j+1,"\t",results,"\n",file=filename,append=T);
		cat(i+1,"\t",j+1,"\t",diff_mpi[results],"\n",file=filename,append=T);
		cat(i+1,"\t",j+1,"\t",zscore_mpi[results],"\n",file=filename,append=T);
		cat(i+1,"\t",j+1,"\t",poverall[results],"\n",file=filename,append=T);
		cat(i+1,"\t",j+1,"\t",n_obs_mpi[results],"\n",file=filename,append=T);
		cat(i+1,"\t",j+1,"\t",zscore_mpi[results[1:10]],'(',length(results),")\n",sep='');   
	

# end - # note which goes by columns first
